@@ -12,6 +12,9 @@ use std::{
     sync::Arc,
 };
 
+#[cfg(feature = "profile")]
+mod profiler;
+
 mod engine;
 use engine::{builders::EngineBuilder, Engine, ReadGuard, ReadGuardFuture, MAX_ALLOCATION};
 
@@ -77,6 +80,9 @@ enum DeviceError {
     MappingFailed,
     Access,
     ShaderCompilationFailed,
+    // Unable to open summary file
+    #[cfg(feature = "profile")]
+    ProfileSummaryError,
 }
 
 type DeviceResult<T> = Result<T, DeviceError>;
@@ -123,7 +129,9 @@ use write_only::WriteOnly;
 #[derive(Debug, Clone)]
 pub struct DeviceInfo {
     name: String,
+    #[allow(unused)]
     vendor: usize,
+    #[allow(unused)]
     device: usize,
     api: Api,
     device_type: DeviceType,
@@ -324,6 +332,7 @@ pub mod builders {
             } else {
                 self.device.replace(device.clone());
             }
+            self.check_args_size(self.args.len() + 1, false)?;
             let declared_mutable = *self.descriptor.buffers.get(self.args.len()).unwrap();
             if !mutable && declared_mutable {
                 bail!(
@@ -333,9 +342,6 @@ pub mod builders {
                     &self.entry,
                 )
             }
-            if cfg!(debug_assertions) {
-                self.check_args_size(self.args.len() + 1, false)?;
-            }
             self.args.push(ComputePassArg { buffer, mutable });
             Ok(ComputePassBuilder {
                 device: self.device,
@@ -479,11 +485,15 @@ pub mod builders {
             self.check_args_size(self.args.len(), true)?;
             self.check_push_constant_size(self.push_constants.len(), true)?;
             //self.check_specialization_size()?;
-            let work_groups = work_groups(global_size, self.descriptor.local_size);
+            let local_size = self.descriptor.local_size;
+            let work_groups = work_groups(global_size, local_size);
             let compute_pass = ComputePass {
                 module: self.module.id,
+                module_name: self.module.name().map_or_else(String::new, str::to_string),
                 entry: self.descriptor.id,
+                entry_name: self.entry,
                 work_groups,
+                local_size,
                 args: self.args,
                 push_constants: self.push_constants,
                 specialization: self.specialization,
@@ -560,8 +570,11 @@ struct ComputePassArg {
 #[derive(Debug)]
 struct ComputePass {
     module: ModuleId,
+    module_name: String,
     entry: EntryId,
+    entry_name: String,
     work_groups: [u32; 3],
+    local_size: [u32; 3],
     args: Vec<ComputePassArg>,
     push_constants: Vec<u8>,
     specialization: Vec<u8>,
