@@ -1,3 +1,5 @@
+#[cfg(feature = "profile")]
+use super::profiler::{ComputePassMetrics, Profiler};
 use super::{
     shader::{EntryDescriptor, EntryId, Module, ModuleId, SPECIALIZATION_SIZE},
     Api, BufferHandle, BufferId, ComputePass, DeviceError, DeviceId, DeviceResult, WriteOnly,
@@ -9,6 +11,10 @@ use crate::{
 use anyhow::{anyhow, bail};
 use crossbeam_channel::{unbounded as unbounded_channel, Receiver, Sender};
 use crossbeam_utils::atomic::AtomicCell;
+#[cfg(feature = "profile")]
+use gfx_hal::query::{
+    CreationError as QueryCreationError, Query, ResultFlags as QueryResultFlags, Type as QueryType,
+};
 use gfx_hal::{
     adapter::{Adapter, DeviceType, MemoryProperties, PhysicalDevice},
     buffer::{CreationError as BufferCreationError, State, SubRange, Usage},
@@ -37,6 +43,8 @@ use parking_lot::{
 };
 #[cfg(windows)]
 use smol::lock::Semaphore;
+#[cfg(feature = "profile")]
+use std::mem::size_of;
 use std::{
     collections::{HashMap, VecDeque},
     convert::TryFrom,
@@ -398,6 +406,14 @@ impl<B: Backend> EngineBase<B> {
         let compute_queue = gpu.queue_groups[0].queues.pop().unwrap();
         let compute_id = compute_family.id();
         let allocator = Allocator::new(&device, &builder.allocator_config)?;
+        // TODO probably convert to using anyhow::Error instead of DeviceError.
+        #[cfg(feature = "profile")]
+        let profiler = Profiler::get()
+            .transpose()
+            .map_err(|_| DeviceError::ProfileSummaryError)?;
+        #[cfg(feature = "profile")]
+        let context = Arc::new(Context::new(device, allocator, adapter, instance, profiler));
+        #[cfg(not(feature = "profile"))]
         let context = Arc::new(Context::new(device, allocator, adapter, instance));
         let (sender, receiver) = unbounded_channel();
         let queue = Queue::new(receiver, context.clone(), compute_queue, compute_id)?;
@@ -581,8 +597,11 @@ engine_methods! {
             }).collect();
         let op = Op::Compute {
             module: compute_pass.module,
+            module_name: compute_pass.module_name,
             entry: compute_pass.entry,
+            entry_name: compute_pass.entry_name,
             work_groups: compute_pass.work_groups,
+            local_size: compute_pass.local_size,
             args,
             push_constants: compute_pass.push_constants,
             specialization: compute_pass.specialization,
@@ -1019,7 +1038,7 @@ impl<B: Backend> Default for MappingChunk<B> {
 
 #[derive(Clone, Copy, Debug)]
 struct HeapInfo {
-    index: usize,
+    // index: usize,
     size: u64,
     id: MemoryTypeId,
 }
@@ -1038,11 +1057,11 @@ impl HeapInfo {
             .map(|(i, heap)| (i, heap.size))
             .collect();
         heaps.sort_by_key(|(_, size)| *size);
-        for (index, size) in heaps {
+        for (_index, size) in heaps {
             for (i, memory_type) in memory_properties.memory_types.iter().enumerate() {
                 if memory_type.properties.contains(properties) {
                     return Some(Self {
-                        index,
+                        // index,
                         size,
                         id: MemoryTypeId(i),
                     });
@@ -1374,7 +1393,7 @@ impl<B: Backend> WriteSlice<B> {
 
 #[derive(Debug)]
 pub(crate) struct WriteGuardBase<B: Backend> {
-    context: Arc<Context<B>>,
+    _context: Arc<Context<B>>,
     slice: WriteSlice<B>,
     drop: bool,
 }
@@ -1387,7 +1406,7 @@ impl<B: Backend> WriteGuardBase<B> {
         len: u32,
     ) -> Self {
         Self {
-            context,
+            _context: context,
             slice: WriteSlice {
                 chunk,
                 barrier: WriteBarrier::default(),
@@ -1573,6 +1592,7 @@ impl<B: Backend> Queue<B> {
         let n_frames = 3;
         let ready_frames = Vec::with_capacity(n_frames);
         let pending_frames = VecDeque::with_capacity(n_frames);
+        let timestamp_period_ns = queue.timestamp_period() as u64;
         let mut queue = Self {
             receiver,
             queue,
@@ -1583,7 +1603,11 @@ impl<B: Backend> Queue<B> {
             modules: HashMap::new(),
         };
         for _ in 0..n_frames {
-            let frame = Frame::new(queue.context.device(), &mut *queue.command_pool)?;
+            let frame = Frame::new(
+                &queue.context,
+                &mut *queue.command_pool,
+                timestamp_period_ns,
+            )?;
             queue.ready_frames.push(frame);
         }
         Ok(queue)
@@ -1610,7 +1634,7 @@ impl<B: Backend> Queue<B> {
     fn poll(&mut self) -> DeviceResult<()> {
         let mut finished = None;
         if let Some(frame) = self.pending_frames.front_mut() {
-            if frame.poll(self.context.device())? {
+            if frame.poll(&self.context)? {
                 finished = self.pending_frames.pop_front();
             }
         }
@@ -1896,8 +1920,11 @@ enum Op<B: Backend> {
     },
     Compute {
         module: ModuleId,
+        module_name: String,
         entry: EntryId,
+        entry_name: String,
         work_groups: [u32; 3],
+        local_size: [u32; 3],
         args: Vec<ComputeArg<B>>,
         push_constants: Vec<u8>,
         specialization: Vec<u8>,
@@ -2050,6 +2077,79 @@ fn barrier_range(offset: u32, mut len: u32) -> SubRange {
     }
 }
 
+#[cfg(feature = "profile")]
+#[derive(Debug)]
+struct QueryPool<B: Backend> {
+    pool: B::QueryPool,
+    buffer: Vec<u32>,
+    period_ns: u64,
+}
+
+#[cfg(feature = "profile")]
+impl<B: Backend> QueryPool<B> {
+    fn new(device: &B::Device, count: u32, period_ns: u64) -> DeviceResult<Self> {
+        let pool = unsafe {
+            device
+                .create_query_pool(QueryType::Timestamp, count)
+                .map_err(|err| match err {
+                    QueryCreationError::OutOfMemory(memory_error) => {
+                        DeviceError::from(memory_error)
+                    }
+                    err => todo!("{:?}", err),
+                })?
+        };
+        Ok(Self {
+            pool,
+            buffer: Vec::new(),
+            period_ns,
+        })
+    }
+    unsafe fn get_compute_pass_timestamps(
+        &mut self,
+        device: &B::Device,
+        metrics: &mut [ComputePassMetrics],
+    ) -> DeviceResult<()> {
+        use gfx_hal::device::WaitError;
+        let timestamps = metrics.len() * 2;
+        let reserve = timestamps
+            .checked_sub(self.buffer.len())
+            .unwrap_or_default();
+        self.buffer.extend(std::iter::repeat(0).take(reserve));
+        let data = bytemuck::cast_slice_mut(&mut self.buffer[0..timestamps]);
+        device
+            .get_query_pool_results(
+                &self.pool,
+                0..timestamps as u32,
+                data,
+                size_of::<u32>() as u32,
+                QueryResultFlags::WAIT,
+            )
+            .map_err(|err| match err {
+                WaitError::OutOfMemory(out_of_memory) => DeviceError::from(out_of_memory),
+                WaitError::DeviceLost(device_lost) => DeviceError::from(device_lost),
+            })?;
+        for (duration, timestamp) in metrics
+            .iter_mut()
+            .flat_map(|metric| [&mut metric.start, &mut metric.end])
+            .zip(self.buffer.iter().copied())
+        {
+            *duration = Duration::from_nanos(timestamp as u64 * self.period_ns)
+        }
+        Ok(())
+    }
+    unsafe fn free(self, device: &B::Device) {
+        device.destroy_query_pool(self.pool);
+    }
+}
+
+#[cfg(feature = "profile")]
+impl<B: Backend> Deref for QueryPool<B> {
+    type Target = B::QueryPool;
+    fn deref(&self) -> &Self::Target {
+        &self.pool
+    }
+}
+
 #[derive(Debug)]
 struct Frame<B: Backend> {
     semaphore: B::Semaphore,
@@ -2060,10 +2160,19 @@ struct Frame<B: Backend> {
     mapping_chunks: HashMap<MappingChunkRef<B>, u32>,
     syncs: Vec<Arc<AtomicBool>>,
     ready_to_submit: bool,
+    #[cfg(feature = "profile")]
+    query_pool: Option<QueryPool<B>>,
+    #[cfg(feature = "profile")]
+    compute_pass_metrics: Vec<ComputePassMetrics>,
 }
 
 impl<B: Backend> Frame<B> {
-    fn new(device: &B::Device, command_pool: &mut B::CommandPool) -> DeviceResult<Self> {
+    fn new(
+        context: &Context<B>,
+        command_pool: &mut B::CommandPool,
+        #[allow(unused)] timestamp_period_ns: u64,
+    ) -> DeviceResult<Self> {
+        let device = context.device();
         let semaphore = device.create_semaphore()?;
         let fence = match device.create_fence(true) {
             Ok(fence) => fence,
@@ -2078,6 +2187,12 @@ impl<B: Backend> Frame<B> {
         unsafe {
             command_buffer.begin_primary(CommandBufferFlags::ONE_TIME_SUBMIT);
         }
+        #[cfg(feature = "profile")]
+        let query_pool = context
+            .profiler
+            .as_ref()
+            .map(|_| QueryPool::new(device, 10_000, timestamp_period_ns))
+            .transpose()?;
         Ok(Self {
             semaphore,
             fence,
@@ -2087,6 +2202,10 @@ impl<B: Backend> Frame<B> {
             mapping_chunks: HashMap::new(),
             syncs: Vec::new(),
             ready_to_submit: false,
+            #[cfg(feature = "profile")]
+            query_pool,
+            #[cfg(feature = "profile")]
+            compute_pass_metrics: Vec::new(),
         })
     }
     fn ready_to_submit(&mut self) -> bool {
@@ -2104,8 +2223,11 @@ impl<B: Backend> Frame<B> {
             }
             Op::Compute {
                 module,
+                module_name,
                 entry,
+                entry_name,
                 work_groups,
+                local_size,
                 args,
                 push_constants,
                 specialization,
@@ -2122,13 +2244,33 @@ impl<B: Backend> Frame<B> {
                 } else {
                     return Ok(Err(Op::Compute {
                         module,
+                        module_name,
                         entry,
+                        entry_name,
                         work_groups,
+                        local_size,
                         args,
                         push_constants,
                         specialization,
                     }));
                 };
+                #[cfg(feature = "profile")]
+                let metric_id = self.compute_pass_metrics.len();
+                #[cfg(feature = "profile")]
+                if self.query_pool.is_some() {
+                    self.compute_pass_metrics.push(ComputePassMetrics {
+                        module_id: module.0,
+                        module_name,
+                        entry_name,
+                        invocations: work_groups
+                            .iter()
+                            .zip(local_size.iter())
+                            .map(|(wg, ls)| *wg as usize * *ls as usize)
+                            .product(),
+                        start: Duration::default(),
+                        end: Duration::default(),
+                    });
+                }
                 let descriptors = args.iter().map(|arg| {
                     Descriptor::Buffer(
                         arg.slice.buffer(),
@@ -2170,7 +2312,6 @@ impl<B: Backend> Frame<B> {
                         if type_eq::<B, DX12>() {
                             states.end |= State::SHADER_WRITE;
                         }
-
                         Barrier::Buffer {
                             states,
                             target: arg.slice.buffer(),
@@ -2184,7 +2325,27 @@ impl<B: Backend> Frame<B> {
                         Dependencies::empty(),
                         barriers,
                     );
+                    #[cfg(feature = "profile")]
+                    if let Some(pool) = self.query_pool.as_deref() {
+                        self.command_buffer.write_timestamp(
+                            PipelineStage::COMPUTE_SHADER,
+                            Query {
+                                pool,
+                                id: (metric_id * 2) as u32,
+                            },
+                        )
+                    }
                     self.command_buffer.dispatch(work_groups);
+                    #[cfg(feature = "profile")]
+                    if let Some(pool) = self.query_pool.as_ref() {
+                        self.command_buffer.write_timestamp(
+                            PipelineStage::COMPUTE_SHADER,
+                            Query {
+                                pool,
+                                id: (metric_id * 2 + 1) as u32,
+                            },
+                        )
+                    }
                 }
             }
             Op::Copy { src, dst } => {
@@ -2314,23 +2475,39 @@ impl<B: Backend> Frame<B> {
         self.ready_to_submit = true;
         Ok(Ok(()))
     }
-    fn poll(&mut self, device: &B::Device) -> DeviceResult<bool> {
-        if unsafe { device.get_fence_status(&self.fence)? } {
-            for (chunk, offset) in self.mapping_chunks.drain() {
-                chunk.blocks.finish(offset);
-            }
-            for (slice, _) in self.writes.drain(..) {
-                slice.chunk.blocks.drop_guard();
-            }
-            for sync in self.syncs.drain(..) {
-                sync.store(true, Ordering::SeqCst);
-            }
+    fn reset(&mut self, #[allow(unused)] context: &Context<B>) -> DeviceResult<()> {
+        for (chunk, offset) in self.mapping_chunks.drain() {
+            chunk.blocks.finish(offset);
+        }
+        for (slice, _) in self.writes.drain(..) {
+            slice.chunk.blocks.drop_guard();
+        }
+        for sync in self.syncs.drain(..) {
+            sync.store(true, Ordering::SeqCst);
+        }
+        #[cfg(feature = "profile")]
+        if let Some((profiler, query_pool)) =
+            context.profiler.as_ref().zip(self.query_pool.as_mut())
+        {
+            let device = context.device();
             unsafe {
-                self.command_buffer.reset(false);
-                self.command_buffer
-                    .begin_primary(CommandBufferFlags::ONE_TIME_SUBMIT);
-                self.descriptor_set_allocator.reset();
+                query_pool.get_compute_pass_timestamps(device, &mut self.compute_pass_metrics)?;
             }
+            for metric in self.compute_pass_metrics.drain(..) {
+                profiler.compute_pass(metric);
+            }
+        }
+        unsafe {
+            self.command_buffer.reset(false);
+            self.command_buffer
+                .begin_primary(CommandBufferFlags::ONE_TIME_SUBMIT);
+            self.descriptor_set_allocator.reset();
+        }
+        Ok(())
+    }
+    fn poll(&mut self, context: &Context<B>) -> DeviceResult<bool> {
+        if unsafe { context.device().get_fence_status(&self.fence)? } {
+            self.reset(context)?;
             Ok(true)
         } else {
             Ok(false)
@@ -2382,6 +2559,10 @@ impl<B: Backend> Frame<B> {
         device.destroy_semaphore(self.semaphore);
         device.destroy_fence(self.fence);
         self.descriptor_set_allocator.free(device);
+        #[cfg(feature = "profile")]
+        if let Some(query_pool) = self.query_pool {
+            query_pool.free(device);
+        }
     }
 }
 
@@ -2392,6 +2573,8 @@ struct Context<B: Backend> {
     shader_semaphore: Semaphore,
     _adapter: Arc<Adapter<B>>,
     _instance: Arc<B::Instance>,
+    #[cfg(feature = "profile")]
+    profiler: Option<Arc<Profiler>>,
 }
 
 impl<B: Backend> Context<B> {
@@ -2400,6 +2583,7 @@ impl<B: Backend> Context<B> {
         allocator: Allocator<B>,
         adapter: Arc<Adapter<B>>,
         instance: Arc<B::Instance>,
+        #[cfg(feature = "profile")] profiler: Option<Arc<Profiler>>,
     ) -> Self {
         Self {
             device,
@@ -2408,6 +2592,8 @@ impl<B: Backend> Context<B> {
             shader_semaphore: Semaphore::new(1),
             _adapter: adapter,
             _instance: instance,
+            #[cfg(feature = "profile")]
+            profiler,
         }
     }
     fn device(&self) -> &B::Device {
